"""Contains metric aggregator callbacks."""
from typing import Dict, Union

from nn_helper.callbacks.base_callback import BaseCallback


class MetricAggregator(BaseCallback):
    """Callback that records metrics over epochs by taking the mean.

    It makes these computed means available to next callbacks. This callback
    is already incorporated in the `fit` function from the `train.py` script.
    """

    def on_epoch_begin(self, batch_index: int, logs: Dict[str, Union[int, float]] = None) -> None:
        """Initialises the metric aggregator for the current epoch.

        Args:
            batch_index (int): The batch number.
            logs (Dict[str, Union[int, float]], optional): The logs. Defaults to None.
        """
        self.batches = 0
        self.totals = {}
        self.metrics = ["loss"] + self.params["metrics"]

    def on_batch_end(self, batch_index: int, logs: Dict[str, Union[int, float]] = None) -> None:
        """Is called on every batch ending to store metrics from logs.

        Multiple losses can be captured from logs. If the loss function returns,
        for example, `loss`, which also returns the loss splitted into `loss_1` and `loss_2`,
        all three losses are stored.

        Args:
            batch_index (int): The batch number.
            logs (Dict[str, Union[int, float]], optional): The logs from the batch. This includes, among others
                the loss values and metrics. Defaults to None.
        """
        logs = logs or {}

        batch_size = logs.get("size", 1)
        self.batches += batch_size

        for key, value in logs.items():
            if key in self.totals:
                self.totals[key] += value * batch_size
            elif key in self.metrics:
                self.totals[key] = value * batch_size

    def on_epoch_end(self, epoch: int, logs: Dict[str, Union[int, float]] = None) -> None:
        """Make the mean of the metrics in the current epoch available for next callbacks.

        Args:
            epoch (int): The epoch number.
            logs (Dict[str, Union[int, float]], optional): The logs. Defaults to None.
        """
        if not logs:
            for key in self.metrics:
                if key in self.totals:
                    # Make value available to next callbacks.
                    logs[key] = self.totals[key] / self.batches



"""Contains logger classes that, for example, print metrics to terminal."""
from typing import Dict, Union

from nn_helper.callbacks.base_callback import BaseCallback


class VerboseLogger(BaseCallback):
    """Callback that prints training information to the terminal.

    For example, epoch number and corresponding average loss.
    """

    def __init__(self, log_interval: int = 1):
        """Initialise instance variables.

        Args:
            log_interval (int, optional): The interval at which metrics
                are printed to terminal during an epoch. Defaults to 1.
        """
        self.log_interval = log_interval

    def on_train_begin(self, logs: Dict[str, Union[int, float]] = None) -> None:
        """Initialises the verbose logger before training.

        Args:
            logs (Dict[str, Union[int, float]], optional): The logs. Defaults to None.
        """
        self.metrics = ["loss"] + self.params["metrics"]
        if self.params["verbose"] > 0:
            print("Begin training...")

    def on_epoch_begin(self, epoch: int, logs: Dict[str, Union[int, float]] = None) -> None:
        """Initialises the verbose logger for the current epoch.

        Args:
            epoch (int): The epoch number.
            logs (Dict[str, Union[int, float]], optional): The logs. Defaults to None.
        """
        self.batch_iterations = 0

    def on_batch_begin(self, batch_index: int, logs: Dict[str, Union[int, float]] = None) -> None:
        """Initialises the verbose logger for the current batch.

        Args:
            batch_index (int): The batch number.
            logs (Dict[str, Union[int, float]], optional): The logs from the batch. Defaults to None.
        """
        if self.params["verbose"] == 2:
            self.batch_iterations += 1
            if self.batch_iterations % self.log_interval == 0:
                print("Batch:", self.batch_iterations)

    def on_batch_end(self, batch_index: int, logs: Dict[str, Union[int, float]] = None) -> None:
        """Is called at the end of every batch in an epoch.

        Args:
            batch_index (int): The batch number.
            logs (Dict[str, Union[int, float]], optional): The logs from the batch. Defaults to None.
        """
        if self.params["verbose"] == 2:
            if self.batch_iterations % self.log_interval == 0:
                for key in self.metrics:
                    if logs.get(key, None):
                        print(f"{key}: {logs[key]}")

    def on_epoch_end(self, epoch: int, logs: Dict[str, Union[int, float]] = None) -> None:
        """Is called at the end of every epoch in training.

        Args:
            epoch (int): The epoch number.
            logs (Dict[str, Union[int, float]], optional): The logs. Defaults to None.
        """
        if self.params["verbose"] > 0:
            print("Epoch:", epoch)
            for key in self.metrics:
                if logs.get(key, None):
                    print(f"{key}: {logs[key]}")

    def on_train_end(self, logs: Dict[str, Union[int, float]] = None) -> None:
        """Is called once at the end of training.

        Args:
            logs (Dict[str, Union[int, float]], optional): The logs. Defaults to None.
        """
        if self.params["verbose"] > 0:
            print("Finished.")



"""Contains base classes for models that must be used in the nn_helper package."""
from abc import ABC, abstractmethod
from typing import Dict, Tuple

import torch


class ModelBase(ABC, torch.nn.Module):
    """A simple model base class."""

    def __init__(self):
        """Initialisation of an instance."""
        super(ModelBase, self).__init__()

    @abstractmethod
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """The forward pass function.

        Args:
            x (torch.Tensor): Training data.

        Returns:
            torch.Tensor: Output of the forward pass.
        """

    def prepare_batch(
        self, batch: Tuple[torch.Tensor, torch.Tensor], device: str = "cpu"
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Prepares the batch before it is forwarded into the model.

        This function can be overriden if different functionality is needed.

        Args:
            batch (Tuple[torch.Tensor, torch.Tensor]): The training batch data.
            device (str, optional): Whether trained on GPU or CPU. Defaults to "cpu".

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: Preprocessed batch.
        """
        return batch[0].to(device), batch[1].to(device)

    def fit_function(
        self, input_batch: Tuple[torch.Tensor, torch.Tensor], loss_fn: torch.nn.Module, optimizer: torch.optim.Optimizer
    ) -> Dict[str, torch.Tensor]:
        """This orchistrates the complete forward and backward pass.

        This function can be overriden if different functionality is needed.

        Args:
            input_batch (Tuple[torch.Tensor, torch.Tensor]): The training data.
            loss_fn (torch.nn.Module): The loss function that should be used.
            optimizer (torch.optim.Optimizer): The optimizer that should be used.

        Returns:
            Dict[str, torch.Tensor]: The loss value(s).
        """
        self.train()
        optimizer.zero_grad()

        outputs = self(input_batch[0])

        losses = loss_fn(input_batch[1], *outputs)

        if isinstance(losses, torch.Tensor):
            losses = {"loss": losses}

        losses["loss"].backward()
        optimizer.step()

        return losses



"""Provides BaseCallback for callbacks and CallbackList functions as wrapper for callbacks."""
from __future__ import annotations

from typing import Any, Dict, List, Union

from nn_helper import ModelBase


class CallbackList:
    """Container abstracting a list of callbacks."""

    def __init__(self, callbacks: List[BaseCallback]):
        """Provide a list of callbacks to be used.

        Args:
            callbacks (List[Callback]): Callback classes that
                inherit the Callback base class.
        """
        self.callbacks = callbacks

    def set_params(self, params: Dict[str, Any]) -> None:
        """Sets the training parameters.

        Args:
            params (Dict[str, Any]): Training parameters.
        """
        for callback in self.callbacks:
            callback.set_params(params)

    def set_model(self, model: ModelBase) -> None:
        """Sets the available models.

        Args:
            model (ModelBase): The model(s).
        """
        for callback in self.callbacks:
            callback.set_model(model)

    def on_train_begin(self, logs: Dict[str, Union[int, float]] = None) -> None:
        """Called at the beginning of training.

        Args:
            logs (Dict[str, Union[int, float]]): dictionary of logs. Defaults to None.
        """
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_train_begin(logs)

    def on_train_end(self, logs: Dict[str, Union[int, float]] = None) -> None:
        """Called at the end of training.

        Args:
            logs (Dict[str, Union[int, float]], optional): Dictionary of logs. Defaults to None.
        """
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_train_end(logs)

    def on_epoch_begin(self, epoch: int, logs: Dict[str, Union[int, float]] = None) -> None:
        """Called at the beginning of an epoch.

        Args:
            epoch (int): Index of epoch.
            logs (Dict[str, Union[int, float]], optional): Dictionary of logs. Defaults to None.
        """
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_epoch_begin(epoch, logs)

    def on_epoch_end(self, epoch: int, logs: Dict[str, Union[int, float]] = None) -> None:
        """Called at the end of an epoch.

        Args:
            epoch (int): Index of epoch.
            logs (Dict[str, Union[int, float]], optional): Dictionary of logs. Defaults to None.
        """
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_epoch_end(epoch, logs)

    def on_batch_begin(self, batch_index: int, logs: Dict[str, Union[int, float]] = None) -> None:
        """Called at the start of a batch.

        Args:
            batch_index (int): Index of batch whithin current epoch.
            logs (Dict[str, Union[int, float]], optional): Dictionary of logs. Defaults to None.
        """
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_batch_begin(batch_index, logs)

    def on_batch_end(self, batch_index: int, logs: Dict[str, Union[int, float]] = None) -> None:
        """Called at the end of a batch.

        Args:
            batch_index (int): Index of batch whithin current epoch.
            logs (Dict[str, Union[int, float]], optional): Dictionary of logs. Defaults to None.
        """
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_batch_end(batch_index, logs)


class BaseCallback:
    """Base class that every Callback must inherit."""

    def __init__(self):
        """Initialise BaseCallback."""
        self.model = None

    def set_params(self, params: Dict[str, Any]) -> None:
        """Set parameters.

        Args:
            params (Dict[str, Any]): Training parameters.
        """
        self.params = params

    def set_model(self, model: ModelBase) -> None:
        """Set model(s).

        Args:
            model (ModelBase): The model(s).
        """
        self.model = model

    def on_epoch_begin(self, epoch: int, logs: Dict[str, Union[int, float]] = None) -> None:
        """Called at the beginning of every epoch.

        Args:
            epoch (int): At which epoch the training loop is.
            logs (Dict[str, Union[int, float]], optional): Dictionary of logs. Defaults to None.
        """

    def on_epoch_end(self, epoch: int, logs: Dict[str, Union[int, float]] = None) -> None:
        """Called at the end of every epoch.

        Args:
            epoch (int): At which epoch the training loop is.
            logs (Dict[str, Union[int, float]], optional): Dictionary of logs. Defaults to None.
        """

    def on_batch_begin(self, batch_index: int, logs: Dict[str, Union[int, float]] = None) -> None:
        """Called at the beginning of every batch.

        Args:
            batch_index (int): At which batch the epoch loop is.
            logs (Dict[str, Union[int, float]], optional): Dictionary of logs. Defaults to None.
        """

    def on_batch_end(self, batch_index: int, logs: Dict[str, Union[int, float]] = None) -> None:
        """Called at the end of every batch.

        Args:
            batch_index (int): At which batch the epoch loop is.
            logs (Dict[str, Union[int, float]], optional): Dictionary of logs. Defaults to None.
        """

    def on_train_begin(self, logs: Dict[str, Union[int, float]] = None) -> None:
        """Called once at the beginning of training.

        Args:
            logs (Dict[str, Union[int, float]], optional): Dictionary of logs. Defaults to None.
        """

    def on_train_end(self, logs: Dict[str, Union[int, float]] = None) -> None:
        """Called once at the end of training.

        Args:
            logs (Dict[str, Union[int, float]], optional): Dictionary of logs. Defaults to None.
        """



"""Contains logic to train Pytorch Neural networks."""
from typing import Any, Dict, List, Tuple

import torch

from nn_helper.base.model import ModelBase
from nn_helper.callbacks.base_callback import BaseCallback, CallbackList
from nn_helper.callbacks.metric_aggregator import MetricAggregator


def fit(
    model: ModelBase,
    data_loader: torch.utils.data.DataLoader,
    epochs: int,
    loss_fn: torch.nn.Module,
    optimizer: torch.optim.Optimizer,
    callbacks: List[BaseCallback] = None,
    metrics: List[str] = [],
    verbose: int = 1,
    device: str = "cpu",
    start_epoch: int = 1,
) -> None:
    """Fits the model on the training data and calls callbacks during training in different phases.

    Args:
        model (ModelBase): The model that should be trained.
        data_loader (torch.utils.data.DataLoader): The training data.
        epochs (int): The number of passes through the training data.
        loss_fn (torch.nn.Module): The loss function.
        optimizer (torch.optim.Optimizer): The optimizer.
        fit_function (Callable, optional): Function that defines how to train
            the model. Defaults to standard_fit_function.
        callbacks (List[BaseCallback], optional): Which functions to call at the start
            or end of every batch and epoch. Defaults to None.
        metrics (List[str], optional): Which metrics that should
            be tracked. Defaults to [].
        verbose (int, optional): [description]. Defaults to 1.
        device (str, optional): Whether training is done on cpu or gpu.
            Defaults to 'cpu'.
        start_epoch (int, optional): Where to start training. This is useful
            when training needs to resume from a certain epoch point.
            Defaults to 1.
    """
    assert isinstance(model, ModelBase), f"Your {str(model)} must be of type {str(ModelBase)}."
    num_batches = len(data_loader)

    batch_size = data_loader.batch_size

    callbacks = CallbackList([MetricAggregator()] + (callbacks or []))
    callbacks.set_model(model)
    callbacks.set_params(
        {
            "optimizer": optimizer,
            "num_batches": num_batches,
            "batch_size": batch_size,
            "metrics": metrics,
            "verbose": verbose,
        }
    )

    callbacks.on_train_begin()

    for epoch in range(start_epoch, epochs + 1):
        callbacks.on_epoch_begin(epochs)

        epoch_logs = {}
        for batch_index, batch in enumerate(data_loader):
            batch_logs = dict(batch=batch_index, size=(batch_size or 1))

            def handle_batch(batch_logs: Dict[str, Any], batch: Tuple[torch.Tensor, torch.Tensor]) -> None:
                """Processes the batch through the neural network.

                Preprocesses the batch, adds it to the device, and applies a forward and a backward pass.

                Args:
                    batch_logs (Dict[str, Any]): The logs.
                    batch (Tuple[torch.Tensor, torch.Tensor]): The data.
                """
                x, y = model.prepare_batch(batch)

                callbacks.on_batch_begin(batch_index, batch_logs)

                losses = model.fit_function((x, y), loss_fn, optimizer)

                for loss in losses:
                    batch_logs[loss] = losses[loss].item()

            handle_batch(batch_logs, batch)

            callbacks.on_batch_end(batch_index, batch_logs)

        callbacks.on_epoch_end(epoch, epoch_logs)

    callbacks.on_train_end()



"""Installation script for the nn_helper package."""
from setuptools import find_packages, setup

VERSION = "0.0.1"
DESCRIPTION = "Neural network helper for Pytorch."
LONG_DESCRIPTION = "Neural network wrapper for Pytorch to easily build and train neural networks."

test = ["coverage>=6.0.2", "pre-commit>=2.15.0"]


setup(
    name="nn_helper",
    version=VERSION,
    license="MIT",
    author="Mark de Blaauw",
    author_email="markdeblaauw@gmail.com",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    url="https://gitlab.com/markdeblaauw/nn-helper",
    packages=find_packages(include=["nn_helper", "nn_helper.*"]),
    install_requires=["torch>=1.9.1", "numpy>=1.21.2"],
    extras_require={
        "dev": test,
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords=["python", "neural network", "pytorch"],
    python_requires=">=3.8",
)
